
#include "esp_camera.h"
#include "Arduino.h"
#include "FS.h"                // SD Card ESP32
#include "SD_MMC.h"            // SD Card ESP32
 
#include "driver/rtc_io.h"


#include "LocalDefine.h"


// ------ DHT sensor
#include "Adafruit_Sensor.h"
#include "DHT.h"    // Released under an MIT license.
#include "DHT_U.h"  // Released under an MIT license.


// ------ WIFI
#include <WiFi.h>
#include <WiFiClient.h>

// ------ TIME rtc
#include "ESP32Time.h"


#include "ESP32_FTPClient.h"

DHT_Unified dht(DHTPIN, DHTTYPE);

ESP32Time rtc(3600);  // offset in seconds GMT+1
 
 

String wifi_ssid; 
String wifi_pass;
String ntpServer;

char ftp_server[25];
char ftp_user[25];
char ftp_pass[25];

String v_year="1900";

char v_tmp_msg[200];


int v_debugger = 1; // Mode proves per defecte 

int v_sleep = 300; // temps sleep per defecte (en segons) 5 min 

char namesFiles[MAX_FILE][LEN_NAME];
String Array_TempHum[C_CAPTURES];
 
ESP32_FTPClient ftpClient (ftp_server,21,ftp_user,ftp_pass, 5000, 2);

void setup() {

  SD_MMC.begin("/sdcard", true);

  Serial.begin(115200);
  //Serial.setDebugOutput(true);

  Load_config();
  Active_Wifi();
  
  v_year=rtc.getTime("%M");
  if(atoi(v_year.c_str())<2022)
    Set_time_prg();

  if(!SD_MMC.begin()){
    MSG("ERROR BEGIN SD");
    return;
  }
  
  uint8_t cardType = SD_MMC.cardType();
  if(cardType == CARD_NONE){
    MSG("ERROR SD CARD_NONE");
    return;
  }
  
  Send_ftp();

  set_camara();

  for (unsigned short i = 0; i < C_CAPTURES; i++) Take_photo();

  for (unsigned short i = 0; i < C_CAPTURES; i++) 
    Take_TempHum(i);    

  Save_Sensor();

  Hibernate();

}  

void MSG( String p_msg ) {
    if (v_debugger) 
          Serial.printf("%s\n", p_msg.c_str());
}

void MSG( String p_msg, int p_valor ) {
    if (v_debugger) 
          Serial.printf("%s %d\n", p_msg.c_str(), p_valor);
}

void Hibernate() {
  MSG("....Hibernate()");
  
  Serial.flush();
  
  //Calculate Time to sleep -> allways wake up in clock hours
  String v_get_min=rtc.getTime("%M");

  int v_next_wake_up = atoi(v_get_min.c_str());
  v_next_wake_up = (61 - v_next_wake_up) * 60 ;  // allways in next hour

  MSG("Calculate next_wake_up = " , v_next_wake_up  );

  if (v_debugger)   
      v_next_wake_up = v_sleep; 

  MSG("Last next_wake_up aplicate = " , v_next_wake_up  );
  
  esp_sleep_enable_timer_wakeup(C_FACTOR_TIME_TO_SLEEP * v_next_wake_up);
  esp_deep_sleep_start();
  
}

void Take_TempHum( unsigned short p_attempt) {
  MSG("....Take_TempHum()");
  dht.begin();
  sensors_event_t event;

  float v_Temp, v_Hum;
 
  delay(C_CAPTURA_DELAY);
  
  for (unsigned short  i = 0; i < C_SENSOR_ATTEMPTS; i++) {
    MSG("Temperature .... opc:", i);

    dht.temperature().getEvent(&event);
    v_Temp = event.temperature;
    delay(C_SENSOR_DELAY);
    if (!isnan(v_Temp)) break;
  }

  delay(C_SENSOR_DELAY);
  for (unsigned short i = 0; i < C_SENSOR_ATTEMPTS; i++) {
    MSG("humidity .... opc:", i);

    dht.humidity().getEvent(&event);
    v_Hum = event.relative_humidity;
    if (!isnan(v_Hum)) break;
    delay(C_SENSOR_DELAY);
  }

  if (isnan(v_Temp)) v_Temp = (float) -100;
  if (isnan(v_Hum)) v_Hum = (float) -100;
  
  char v_reg[100]; 
  String v_time_reg= String(rtc.getTime("%Y%m%d%H%M%S"));

  sprintf(v_reg, "%s;%f;%f\n", v_time_reg.c_str() , v_Temp, v_Hum);

  MSG("Register Temp/Hum : ");
  MSG(v_reg);
  

  Array_TempHum[p_attempt]=v_reg;
 
  MSG("===================");
  
}

void Save_Sensor() {

  MSG("....Save_Sensor()");

  char v_name_file[100] = "/TempHum.csv";

  File file;
  fs::FS &fs = SD_MMC; 
  
  for (unsigned short i = 0; i < C_FILE_ATTEMPTS; i++) {
    MSG("Open file TempHum opc:", i);    
    file.close();
    file = fs.open(v_name_file, FILE_APPEND);

    if (file) break;
    MSG("ERROR OPEN TempHum.csv");

    delay(C_GENERAL_DELAY);
    
  }

  if (!file) return;

  for  (unsigned short i=0; i<C_CAPTURES ; i++){
      if (Array_TempHum[i]!=NULL)
        file.print(Array_TempHum[i]);

      Array_TempHum[i]="";

  }

  file.close();

  MSG("===================");
 
}

void set_camara() {
  MSG("....set_camara()");

  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG; 
      
  config.fb_location = CAMERA_FB_IN_DRAM; // CAMERA_FB_IN_DRAM; // CAMERA_FB_IN_PSRAM  CAMERA_FB_IN_DRAM

  config.frame_size = FRAMESIZE_QVGA; // FRAMESIZE_ + QVGA|CIF|VGA|SVGA|XGA|SXGA|UXGA
  config.jpeg_quality = 12;
  config.fb_count = 1;


  // Init Camera
  esp_err_t err;
  for (unsigned short i = 0; i < C_PHOTO_ATTEMPTS; i++) {
    err = esp_camera_init(&config);
    if (err != ESP_OK) {
      MSG("ERROR esp_camera_init");
    }
    else 
      break;

    delay(C_GENERAL_DELAY);

  }  



  sensor_t * s = esp_camera_sensor_get();

  // PIXFORMAT_JPEG
   //s->set_framesize(s, (framesize_t)4);   // framesize
   s->set_quality(s, 4);               //quality 6 -> 12 
   s->set_contrast(s, 2);                //contrast     -2 to 2
   s->set_brightness(s, -2);              //brightness   -2 to 2
   s->set_saturation(s, -2);              //saturation   -2 to 2
   s->set_colorbar(s, 0);                //colorbar     0 = disable , 1 = enable
   s->set_whitebal(s, 0);                //awb          0 = disable , 1 = enable
   s->set_gain_ctrl(s, 0);               //agc          0 = disable , 1 = enable
   s->set_exposure_ctrl(s, 0);           //aec          0 = disable , 1 = enable
   s->set_hmirror(s, 0);                 //hmirror      0 = disable , 1 = enable
   s->set_vflip(s, 0);                   //vflip        0 = disable , 1 = enable
   s->set_awb_gain(s, 30);                //awb_gain     0 to 30
   s->set_agc_gain(s, 0);                //agc_gain     0 = disable , 1 = enable
   s->set_aec_value(s, 1200);             //aec_value    0 to 1200
   s->set_aec2(s, 0);                    //aec2         0 = disable , 1 = enable
   s->set_bpc(s, 0);                     //bpc          0 = disable , 1 = enable
   s->set_wpc(s, 0);                     //wpc          0 = disable , 1 = enable
   s->set_raw_gma(s, 1);                 //raw_gma      0 = disable , 1 = enable
   s->set_lenc(s, 0);                    //lenc         0 = disable , 1 = enable
   s->set_wb_mode(s, 0);                 //wb_mode      0 to 4 - if awb_gain enabled (0 - Auto, 1 - Sunny, 2 - Cloudy, 3 - Office, 4 - Home
   s->set_ae_level(s, -2);               //ae_level     -2 to 2
   s->set_special_effect(s, 1);          //Effect       0 to 6 (0 - No Effect, 1 - Negative, 2 - Grayscale, 3 - Red Tint, 4 - Green Tint, 5 - Blue Tint, 6 - Sepia)
   s->set_dcw(s, 0);                     //dcw          0 = disable , 1 = enable
   s->set_gainceiling(s, (gainceiling_t)0);  //gainceiling  0 to 6 

  MSG("===================");

}  



void Take_photo() {

  MSG("....Take_photo()");

  delay(C_CAPTURA_DELAY);

  camera_fb_t * fb = NULL;
  
  pinMode(GPIO_NUM_4, OUTPUT); // Flash  
  digitalWrite(GPIO_NUM_4, HIGH); 
  rtc_gpio_hold_en(GPIO_NUM_4);  
 
  // Take Picture with Camera
  for (unsigned short i = 0; i < C_PHOTO_ATTEMPTS; i++) {    

    MSG("Take Picture opc:", i);    

    delay(C_PHOTO_DELAY);  
    fb= esp_camera_fb_get();

    if (fb)
      break;
    else{
          esp_camera_fb_return(fb);
          MSG("ERROR esp_camera_fb_get()");   
          fb = NULL;    
    }

  }

  // Turns off the ESP32-CAM white on-board LED (flash) connected to GPIO 4
  digitalWrite(GPIO_NUM_4, LOW);
  rtc_gpio_hold_dis(GPIO_NUM_4);
   

  // Path where new picture will be saved in SD Card
  String path = "/" + String(rtc.getTime("%Y%m%d%H%M%S")) +".jpg";

  fs::FS &fs = SD_MMC; 
 
  File file;
  for (unsigned short i = 0; i < C_PHOTO_ATTEMPTS; i++) {
    MSG("Save photo in SD opc:", i);    

    file.close();
    file = fs.open(path.c_str(), FILE_WRITE);
    if (file) break;
    delay(C_GENERAL_DELAY);
    
  }

  if (!file){
    MSG("ERROR FILE");   
     return;   
  }

  file.write( fb->buf, fb->len );


  file.close();
  
  esp_camera_fb_return(fb); 
  

  
 
  MSG("===================");

}


void Load_config() { 

  MSG("....Load_config()"); 

  File file;
  fs::FS& fs = SD_MMC;

  for (unsigned short i = 0; i < C_SD_ATTEMPTS; i++) {
    MSG("Open File config:", i);     
    file = SD_MMC.open(C_FILE_CONFIG, FILE_READ);
    if (file) break;
    file.close();
    delay(C_SD_DELAY);
  }

  String v_line, v_item, v_value;
  int v_pos;
  while (file.available()) {
      
      v_line = file.readStringUntil('\n'); 
      
      int v_pos = v_line.indexOf("=");

      if (v_pos < 1){
          file.close();
          break;
      } 
      
      v_item = v_line.substring(0,v_pos); 
      v_value = v_line.substring(v_pos+1); 
      
      if(v_item == "WIFI_SSID")
          wifi_ssid=v_value; 
      if(v_item == "WIFI_PASS")
          wifi_pass=v_value;
      if(v_item == "NTPSERVER")
          ntpServer=v_value; 
      if(v_item == "FTP_SERVER")
          strcpy(ftp_server,v_value.c_str()); 
      if(v_item == "FTP_USER")
          strcpy(ftp_user,v_value.c_str()); 
      if(v_item == "FTP_PASS")
          strcpy(ftp_pass,v_value.c_str()); 
          
      if(v_item == "DEBUGGER" && v_value!= "1") 
          v_debugger=0; 

      if(v_item == "SLEEP" && !v_value.isEmpty() ) 
          v_sleep = v_value.toInt();
          
  }

  file.close();

 // Without parameters, the data taken does not have a reference either, so it is better to download them and that is why it executes hibernate
  if (wifi_ssid.isEmpty() || wifi_pass.isEmpty() || ntpServer.isEmpty() || ftp_server[0]==0 || ftp_user[0]==0 || ftp_pass[0]==0 ) { 
      MSG("ERROR LOAD config");  
      Hibernate();
  }

  MSG("===================");

}
 
void Active_Wifi () {

  MSG("....Active_Wifi()");
  
  WiFi.begin( wifi_ssid.c_str(), wifi_pass.c_str() );

  for (unsigned short i = 0; i < C_WIFI_ATTEMPTS; i++) {
    if (WiFi.status() == WL_CONNECTED) 
        break;
    else
        delay(C_WIFI_DELAY);

  }
  
  if (WiFi.status() != WL_CONNECTED) {
   MSG("ERROR CONNECTED WIFI"); 
   Hibernate();

  }
  
  MSG("===================");

}


void Set_time_prg() {

  MSG("....Set_time_prg()");
  
  
  long gmtOffset_sec = 3600;
  int daylightOffset_sec = 3600;
   
  //init and get the time
  
  setenv("TZ", "CET-1CEST-2,M3.5.0/02:00:00,M10.5.0/03:00:00", 1);

  tzset();

  struct tm timeinfo;

  for (unsigned short i = 0; i < C_TIME_ATTEMPTS; i++) {
    configTime(gmtOffset_sec, daylightOffset_sec, ntpServer.c_str());
    if (!getLocalTime(&timeinfo)) 
        MSG("ERROR GET TIME opc-> ", i);
    else 
      break;  

    delay(C_GENERAL_DELAY);
      
  }
  
  if (!getLocalTime(&timeinfo)) 
    Hibernate();

  MSG("===================");
 
}


void Send_ftp() {

 
  MSG("....Send_ftp()");


  for (unsigned short i = 0; i < C_WIFI_ATTEMPTS; i++) {
    MSG("Reconnect wifi opc-> ", i);     
    if(!WiFi.isConnected()) 
      delay(C_WIFI_DELAY);
    else 
      break;

    WiFi.reconnect();

  }
 
  unsigned int n_files;
  n_files = printDirectory("/");


  for (unsigned short i = 0; i < C_FTP_ATTEMPTS; i++) {
    MSG("Open FTP opc-> ", i);     

    delay(C_FTP_DELAY);      
    ftpClient.OpenConnection();  
    
    if(ftpClient.isConnected()) 
      break;
  }

  if(!ftpClient.isConnected()) {
    MSG("ERROR CONNECTED FTP");
    return;
  }


  for ( unsigned int i= 0; i< n_files; i++)
    File_to_ftp(namesFiles[i]);
 
  ftpClient.CloseConnection();
 
  MSG("===================");
 
}


unsigned int printDirectory(char* root) {

  MSG("....printDirectory()");
  
  File dir = SD_MMC.open(root);

  char Filename[LEN_NAME];

  unsigned int n_files=0;

  while(true) {
    File entry =  dir.openNextFile();
    if (!entry && n_files < MAX_FILE ) 
        break;
    
    namesFiles[n_files][0]=0;
    strcpy(Filename,entry.name());

    if (!entry.isDirectory() && isFilterFile(Filename) ) {
          strcpy(namesFiles[n_files],Filename);   
          n_files++;
    }        

    entry.close();
    if (n_files==MAX_FILE) break;

  }

  MSG("===================");

  return n_files;
}


bool isFilterFile(char* filename) {

  MSG("....isFilterFile()");
  
  int8_t len = strlen(filename);
  bool result;
  if (  strstr(strlwr(filename + (len - 4)), ".jpg")
     || strstr(strlwr(filename + (len - 5)), ".jpeg")
     || strstr(strlwr(filename + (len - 4)), ".bmp")
     || strstr(strlwr(filename + (len - 4)), ".gif")
     || strstr(filename, "TempHum.csv") 
    ) {
    result = true;
  } else {
    result = false;
  }

  MSG("===================");
  
  return result;
}


void File_to_ftp(char* filename) {

  MSG("....File_to_ftp():");
  MSG(filename);   

   fs::FS &fs = SD_MMC; 

 //ftpClient.ChangeWorkDir("my_new_dir");

  String ftp_file;

  if(strstr(filename, "TempHum.csv")){
    ftpClient.InitFile("Type A");
    ftp_file="TempHum_";
    ftp_file.concat(String(rtc.getTime("%Y%m%d%H%M%S")));
    ftp_file.concat(".csv");
  }
  else{
    ftpClient.InitFile("Type I"); 
    ftp_file=filename;
  }

  ftpClient.NewFile(ftp_file.c_str());

  String fullPath = "/";
  fullPath.concat(filename);

  File file = fs.open(fullPath.c_str());
  if (!file) {
    MSG("ERROR open file for reading");
    return;
  }


  while (file.available()) {
    // Create and fill a buffer
    unsigned char buf[5120];
    size_t readVal = file.read(buf, sizeof(buf));
    ftpClient.WriteData(buf, readVal);
  }

  ftpClient.CloseFile();

  file.close();

  char v_exist[100];

  
  //ftpClient.OpenConnection();  
  
  ftpClient.GetLastModifiedTime(ftp_file.c_str(), v_exist);

  if(v_exist[0]!=0) // When  File Exists in Ftp delete from SD
    fs.remove(fullPath.c_str()); // 

  MSG("===================");

  
}


void loop() {
  
}
